unit EVPmain;

interface

uses
    Windows,
    Messages,
    SysUtils,
    Variants,
    Classes,
    Graphics,
    Controls,
    Forms,
    Dialogs,
    Menus,
    PluginCore,
    F8Utils,
    F8GLUtils,
    F8OpenGL,
    GL,
    F8GLText,
    F8PaintBox,
    VCL.StdCtrls,
    VCL.ExtCtrls;


type

    FrontCarInfoType = record
        Exists      : Boolean;
        Distance    : Double;

        end;

    TSamplePlugin = class(TF8PluginClass, IF8OpenGLPlugin)
        strict private
            p_winRoadApplication : IF8ApplicationServices;
            p_Road      : IF8Road;
            p_PaintBox  : TF8Paintbox;
            p_Button    : TButton;
            p_Edit      : TEdit;
            p_Interval  : Double;
            p_CarInstance  :IF8CarInstance;
            p_Ribbon    : IF8Ribbon;
            p_RibbonTab : IF8RibbonTab;
            p_RibbonGroup : IF8RibbonGroup;
        public
            procedure AfterConstruction; override;
            procedure BeforeDestruction; override;

            //  Event Handler


            procedure EditChange(Sender: TObject);

            //  Implement IF8OpenGLPlugin interface...
            procedure   PaintScene(const opengl : TF8OpenGL; const view: ViewType);
            procedure   PaintHUD(const opengl : TF8OpenGL; const view: ViewType);
            procedure   SetupCamera(const opengl : TF8OpenGL; const view: ViewType);
            procedure   CreateRibbons;
            procedure   FreeRibbons;
            procedure   ProjectDestroyProc;
            procedure   RemoveForms;
            procedure   AbleMenus(enable: Boolean);
            procedure   LoadSignalsControllerClick(Sender: TObject);

            function    ChangeLane_rightthree( const aDriver: IF8MainDriver ): FrontCarInfoType;
            function    ChangeLane_righttwo( const aDriver: IF8MainDriver ): FrontCarInfoType;
            function    ChangeLane_leftthree( const aDriver: IF8MainDriver ): FrontCarInfoType;
            function    ChangeLane_lefttwo( const aDriver: IF8MainDriver ): FrontCarInfoType;



            procedure   DrawInfo( const Value: FrontCarInfoType );
            property    roadAPI : IF8ApplicationServices read p_winRoadApplication;

    end;
              procedure LoadPlugin;
            procedure UnloadPlugin;


//UC-win/Road plugin callback
//procedure RegisterUserPlugin(out optionName, apiVersion : String);                    //before v10.2.0 ,since v11.2.0
procedure RegisterUserPlugin(out optionName, apiVersion, copyright: String);            //since  v10.2.0

var
    samplePlugin : TsamplePlugin;

implementation

uses
   EVPtoolbar,
   EVPsettings;

procedure LoadPlugin;
    begin
    if not Assigned(samplePlugin) then
        samplePlugin := TsamplePlugin.Create;
    if Assigned(samplePlugin) and Assigned(samplePlugin.roadAPI) then
        samplePlugin.roadAPI.RegisterPluginObject(samplePlugin);
    end;

procedure UnloadPlugin;
    begin
    if Assigned(samplePlugin) and Assigned(samplePlugin.roadAPI) then
        samplePlugin.roadAPI.UnRegisterPluginObject(samplePlugin);
    samplePlugin.Free();
    end;

procedure TsamplePlugin.AfterConstruction;
     var
        method : TMethod;

    begin
    inherited;
    Supports(ApplicationServices, IF8ApplicationServices, p_winRoadApplication);
    CreateRibbons;

    PluginAbleMenusProc(method) := AbleMenus;
    p_winRoadApplication.RegisterEventHandler(_plgPluginAbleMenus, method);

    PluginNotifyProc(method):= ProjectDestroyProc;
    p_winRoadApplication.RegisterEventHandler(_plgProjectDestroy,method);


    end;

// ====================================================================
// On plugin destruction :
// Releases the interface of the application.
// ====================================================================
procedure TsamplePlugin.BeforeDestruction;
    var
        method : TMethod;
    begin
    FreeRibbons;

    PluginAbleMenusProc(method) := AbleMenus;
    p_winRoadApplication.UnRegisterEventHandler(_plgPluginAbleMenus, method);

    PluginNotifyProc(method):= ProjectDestroyProc;
    p_winRoadApplication.UnRegisterEventHandler(_plgProjectDestroy,method);

    RemoveForms;
    p_winRoadApplication := nil;
    inherited;
    end;
procedure TsamplePlugin.ProjectDestroyProc;
    begin
    RemoveForms;
    end;

procedure TsamplePlugin.RemoveForms;
    begin
    if Assigned(EmergencyVehicles) then
        FreeAndNil(EmergencyVehicles);
    end;

procedure TsamplePlugin.CreateRibbons;
  const
  PLUGIN_SAMPLE_TABCONTROLNAME = 'Plugin';
  PLUGIN_SAMPLE_TABCAPTION = 'Emergency Vehicle';
  PLUGIN_SAMPLE_RIBBONGROUP_CONTROLNAME = 'Settings';
  PLUGIN_SAMPLE_RIBBONGROUP_NAME = 'Emergency Vehicle Plugin';
    begin
    if Assigned(p_winRoadApplication) and not Assigned(p_Ribbon) then
        begin
        p_Ribbon := p_winRoadApplication.mainForm.GetMainRibbonMenu;
        p_RibbonTab := p_winRoadApplication.mainForm.GetMainRibbonMenuTabByName(PLUGIN_SAMPLE_TABCONTROLNAME);
        if Assigned(p_Ribbon) and not Assigned(p_RibbonTab)then
            begin
            p_RibbonTab := p_Ribbon.CreateRibbonTab(PLUGIN_SAMPLE_TABCONTROLNAME,10000);
            p_RibbonTab.Caption := PLUGIN_SAMPLE_TABCAPTION;
            end;
        end;
    p_RibbonGroup := p_RibbonTab.CreateRibbonGroup(PLUGIN_SAMPLE_RIBBONGROUP_CONTROLNAME,1);
    p_RibbonGroup.Caption := PLUGIN_SAMPLE_RIBBONGROUP_NAME;

    if not assigned(Form1) then
        Form1 := TForm1.Create(nil);
    Form1.Button1.OnClick := LoadSignalsControllerClick;
    if Assigned(p_RibbonGroup) then
        begin
        p_RibbonGroup.Width := Form1.Button1.Width;
        p_RibbonGroup.AddGroupControl(Form1.Button1);
        end;
    p_Ribbon := nil;
    end;
procedure TsamplePlugin.FreeRibbons;
    begin
    //Remove Control from the RibbonGroup
    if Assigned(Form1) and Assigned(p_RibbonGroup) then
        begin
        p_RibbonGroup.RemoveGroupControl(Form1.Button1);
        FreeAndNil(Form1);
        end;
    //Remove RibbonGroup from the RibbonTab
    if Assigned(p_RibbonTab) then
        begin
        p_RibbonTab.DeleteGroup(p_RibbonGroup);
        p_RibbonGroup := nil;
        end;
    p_RibbonTab := nil;
    p_Ribbon := nil;
    end;

//==============================================================================
//  Called by UC-win/Road to enable menus.
//==============================================================================
procedure TsamplePlugin.AbleMenus(enable: Boolean);
    begin
    if Form1 = nil then
       if Form1.Button1 = nil then
         Form1.Button1.Enabled := enable;
    end;

//==============================================================================
//  ClickEvent of the menu
//==============================================================================
procedure TsamplePlugin.LoadSignalsControllerClick(Sender: TObject);
    begin
    if not Assigned(EmergencyVehicles) then
        EmergencyVehicles := TEmergencyVehicles.Create(nil);
        EmergencyVehicles.Show;
    end;


procedure RegisterUserPlugin(out optionName, apiVersion, copyright: String);
    begin
    optionName := 'EmergencyVehiclePlugin';
    apiVersion := PLUGIN_VERSION;
    copyright := 'Brochat';
    end;



procedure TsamplePlugin.EditChange(Sender: TObject);
    begin
    if (p_Edit.Text = '') then
        Exit;

    p_Interval := StrToFloat(p_Edit.Text);
    end;

 // ============================================================================
 // Suitable for different lines.
 // ============================================================================
procedure TsamplePlugin.PaintHUD(const opengl : TF8OpenGL; const view: ViewType);
    var
        Info: FrontCarInfoType;
        lanecount: integer;
        drivingrule: integer;
    begin
    lanecount:=EVPsettings.lanecount;
    drivingrule:=EVPsettings.drivingrule;
    if((lanecount=3) and (drivingrule=0)) then
      Info := ChangeLane_rightthree(p_winRoadApplication.mainForm.driver)
    else if((lanecount=2) and (drivingrule=0)) then
      Info := ChangeLane_righttwo(p_winRoadApplication.mainForm.driver)
    else if((lanecount=3) and (drivingrule=1)) then
      Info := ChangeLane_leftthree(p_winRoadApplication.mainForm.driver)
    else if((lanecount=2) and (drivingrule=1)) then
      Info := ChangeLane_lefttwo(p_winRoadApplication.mainForm.driver);


    DrawInfo(Info);
    end;

procedure TsamplePlugin.PaintScene(const opengl : TF8OpenGL; const view: ViewType);
    begin
    end;

procedure TsamplePlugin.SetupCamera(const opengl : TF8OpenGL; const view: ViewType);
    begin
    end;

procedure TsamplePlugin.DrawInfo(const Value: FrontCarInfoType);
    const
        FONT_SIZE: array [Boolean] of Integer = (24, 40);
        FONT_COLOR: array [Boolean] of TColor = (clWhite, clRed);
    var
        cw: Integer;
        ch: Integer;
        WarningDistance:integer;
        ChangingLaneDistance:integer;
    begin
    // Obtain the size of screen.
    cw := p_winRoadApplication.mainForm.openGL.Width;
    ch := p_winRoadApplication.mainForm.openGL.Height;
    WarningDistance:=EVPsettings.WarningDistance;
    ChangingLaneDistance:=EVPsettings.ChangingLaneDistance;

    F8TextResetSettings;

    F8Text2DModeBegin(cw, ch);

    // Set text position.
    F8TextSetAlignment(_alignCenter);
    F8TextStartFrom(0.5 * cw, 0.8*ch );
    F8TextSetFontSize(30);
    // Show information.
    if Value.Exists then
        begin
        F8TextOut(Format('Distance%4.1f m', [Value.Distance]));
        end
    else
        begin
        F8TextOut('Distance---.- m');
        end;

    F8TextSetAlignment(_alignCenter);
    F8TextStartFrom(0.5 * cw, 0.9*ch );
    F8TextSetFontSize(30);

    if((Value.Distance<ChangingLaneDistance)and (Value.Distance>0)) then
        begin
         F8TextSetFontName('MSPå');
         F8TextSetFontColor(FONT_COLOR[TRUE]);
         F8TextSetFontSize(FONT_SIZE[TRUE]);
         F8TextOut('Sending a lane change request.');
         F8TextLineBreak;
        end

    else if((Value.Distance<WarningDistance) and (Value.Distance>ChangingLaneDistance))  then
        begin
         F8TextSetFontName('MSPå');
         F8TextSetFontColor(FONT_COLOR[FALSE]);
         F8TextSetFontSize(FONT_SIZE[FALSE]);
         F8TextOut('Emergency Vehicle Approching.');
         F8TextLineBreak;
        end
    else
        begin
        F8TextOut(' ');
        end;

    F8Text2DModeEnd;

    glEnable(GL_LIGHTING);
    glDisable(GL_COLOR_MATERIAL);
    end;
// =============================================================================
// 3 lanes, right-hand-drive
// =============================================================================
function TsamplePlugin.ChangeLane_rightthree(const aDriver: IF8MainDriver): FrontCarInfoType;
    var
        p1: TPoint3D;
        p2: TPoint3D;
        number0:Integer;
        number1:Integer;
        aFrontCar: If8CarInstance;
        ChangingLaneDistance:integer;
    begin
     ChangingLaneDistance:=EVPsettings.ChangingLaneDistance;
     // Obtain the information of front car.
    if Assigned(aDriver) and Assigned(aDriver.currentCar) then
        begin
        aFrontCar := aDriver.currentCar.GetFrontCar;
        end
    else
        aFrontCar := NIL;


    if Assigned(aFrontCar) then
        begin
        // Position of emergency vehicle.
        p1.X := aDriver.currentCar.position[_x];
        p1.Y := aDriver.currentCar.position[_y];
        p1.Z := aDriver.currentCar.position[_z];

        // Position of front car.
        p2.X := aFrontCar.position[_x];
        p2.Y := aFrontCar.position[_y];
        p2.Z := aFrontCar.position[_z];

        // Get distance.
        Result.Exists := TRUE;
        Result.Distance := PointDistance(p1, p2);
        number1:= aFrontCar.laneNumber;
        number0:= aDriver.currentCar.laneNumber;
        // Rules of changing lanes.
        if number0 = number1 then
          begin
          if((Result.Distance<ChangingLaneDistance)and (Result.Distance>0)) then
              if number1 = 1 then
              begin
                aFrontCar.ChangeLane(_vclToLeft,10,FALSE)
              end
              else if number1 = 2 then
              begin
                if random > 0.5 then
                    aFrontCar.ChangeLane(_vclToRight,10,FALSE)
                else
                    aFrontCar.ChangeLane(_vclToLeft,10,FALSE)
              end
              else
                aFrontCar.ChangeLane(_vclToRight,10,FALSE)
          end
        // Front cars return to its former lane.
        else
          begin
          if((Result.Distance < ChangingLaneDistance)and (Result.Distance>0)) then
              if number1 = 1 then
              begin
                aFrontCar.ChangeLane(_vclToLeft,200,FALSE)
              end
              else if number1 = 2 then
              begin
                if random > 0.5 then
                    aFrontCar.ChangeLane(_vclToRight,200,FALSE)
                else
                    aFrontCar.ChangeLane(_vclToLeft,200,FALSE)
              end
              else
                aFrontCar.ChangeLane(_vclToRight,200,FALSE)
          end
        end

        // No cars front.
    else
        begin
        Result.Exists := FALSE;
        Result.Distance := 0.0;
        end;
    end;
// =============================================================================
// 2 lanes, right-hand-drive
// =============================================================================
function TsamplePlugin.ChangeLane_righttwo(const aDriver: IF8MainDriver): FrontCarInfoType;
    var
        p1: TPoint3D;
        p2: TPoint3D;
        number0:Integer;
        number1:Integer;
        aFrontCar: If8CarInstance;
        ChangingLaneDistance:integer;
    begin
     ChangingLaneDistance:=EVPsettings.ChangingLaneDistance;
     // Obtain the information of front car.
    if Assigned(aDriver) and Assigned(aDriver.currentCar) then
        begin
        aFrontCar := aDriver.currentCar.GetFrontCar;
        end
    else
        aFrontCar := NIL;

    if Assigned(aFrontCar) then
        begin
        // Position of emergency vehicle.
        p1.X := aDriver.currentCar.position[_x];
        p1.Y := aDriver.currentCar.position[_y];
        p1.Z := aDriver.currentCar.position[_z];
        // Position of front car.
        p2.X := aFrontCar.position[_x];
        p2.Y := aFrontCar.position[_y];
        p2.Z := aFrontCar.position[_z];
        // Get distance.
        Result.Exists := TRUE;
        Result.Distance := PointDistance(p1, p2);
        number1:= aFrontCar.laneNumber;
        number0:= aDriver.currentCar.laneNumber;
        // Rules of changing lanes.
        if number0 = number1 then
          begin
          if((Result.Distance<ChangingLaneDistance)and (Result.Distance>0)) then
              if number1 = 1 then
              begin
                aFrontCar.ChangeLane(_vclToLeft,10,FALSE)
              end
          else
              begin
                aFrontCar.ChangeLane(_vclToRight,10,FALSE)
              end

          end
        // Front cars return to its former lane.
        else
          begin
          if((Result.Distance < ChangingLaneDistance)and (Result.Distance>0)) then
              if number1 = 1 then
              begin
                aFrontCar.ChangeLane(_vclToLeft,200,FALSE)
              end
              else
                aFrontCar.ChangeLane(_vclToRight,200,FALSE)
          end
        end

        // No cars front.
    else
        begin
        Result.Exists := FALSE;
        Result.Distance := 0.0;
        end;
    end;
// =============================================================================
// 3 lanes, left-hand-drive
// =============================================================================
function TsamplePlugin.ChangeLane_leftthree(const aDriver: IF8MainDriver): FrontCarInfoType;
    var
        p1: TPoint3D;
        p2: TPoint3D;
        number0:Integer;
        number1:Integer;
        aFrontCar: If8CarInstance;
        ChangingLaneDistance:integer;
    begin
     ChangingLaneDistance:=EVPsettings.ChangingLaneDistance;
     // Obtain the information of front car.
    if Assigned(aDriver) and Assigned(aDriver.currentCar) then
        begin
        aFrontCar := aDriver.currentCar.GetFrontCar;
        end
    else
        aFrontCar := NIL;
    if Assigned(aFrontCar) then
        begin
        // Position of emergency vehicle.
        p1.X := aDriver.currentCar.position[_x];
        p1.Y := aDriver.currentCar.position[_y];
        p1.Z := aDriver.currentCar.position[_z];

        // Position of front car.
        p2.X := aFrontCar.position[_x];
        p2.Y := aFrontCar.position[_y];
        p2.Z := aFrontCar.position[_z];

        // Get distance.
        Result.Exists := TRUE;
        Result.Distance := PointDistance(p1, p2);
        number1:= aFrontCar.laneNumber;
        number0:= aDriver.currentCar.laneNumber;


        // Rules of changing lanes.
        if number0 = number1 then
          begin
          if((Result.Distance<ChangingLaneDistance)and (Result.Distance>0)) then
              if number1 = 3 then
              begin
                aFrontCar.ChangeLane(_vclToLeft,10,FALSE)
              end
              else if number1 = 2 then
              begin
                if random > 0.5 then
                    aFrontCar.ChangeLane(_vclToRight,10,FALSE)
                else
                    aFrontCar.ChangeLane(_vclToLeft,10,FALSE)
              end
              else
                aFrontCar.ChangeLane(_vclToRight,10,FALSE)
          end
        // Front cars return to its former lane.
        else
          begin
          if((Result.Distance < ChangingLaneDistance)and (Result.Distance>0)) then
              if number1 = 3 then
              begin
                aFrontCar.ChangeLane(_vclToLeft,200,FALSE)
              end
              else if number1 = 2 then
              begin
                if random > 0.5 then
                    aFrontCar.ChangeLane(_vclToRight,200,FALSE)
                else
                    aFrontCar.ChangeLane(_vclToLeft,200,FALSE)
              end
              else
                aFrontCar.ChangeLane(_vclToRight,200,FALSE)
          end
        end

        // No cars front.
    else
        begin
        Result.Exists := FALSE;
        Result.Distance := 0.0;
        end;
    end;
// =============================================================================
// 2 lanes, left-hand-drive
// =============================================================================
function TsamplePlugin.ChangeLane_lefttwo(const aDriver: IF8MainDriver): FrontCarInfoType;
    var
        p1: TPoint3D;
        p2: TPoint3D;
        number0:Integer;
        number1:Integer;
        aFrontCar: If8CarInstance;
        ChangingLaneDistance:integer;
    begin
     ChangingLaneDistance:=EVPsettings.ChangingLaneDistance;
     // Obtain the information of front car.
    if Assigned(aDriver) and Assigned(aDriver.currentCar) then
        begin
        aFrontCar := aDriver.currentCar.GetFrontCar;
        end
    else
        aFrontCar := NIL;
    if Assigned(aFrontCar) then
        begin
        // Position of emergency vehicle.
        p1.X := aDriver.currentCar.position[_x];
        p1.Y := aDriver.currentCar.position[_y];
        p1.Z := aDriver.currentCar.position[_z];

        // Position of front car.
        p2.X := aFrontCar.position[_x];
        p2.Y := aFrontCar.position[_y];
        p2.Z := aFrontCar.position[_z];

        // Get distance.
        Result.Exists := TRUE;
        Result.Distance := PointDistance(p1, p2);
        number1:= aFrontCar.laneNumber;
        number0:= aDriver.currentCar.laneNumber;
        // Rules of changing lanes.
        if number0 = number1 then
          begin
          if((Result.Distance<ChangingLaneDistance)and (Result.Distance>0)) then
              if number1 = 2 then
              begin
                aFrontCar.ChangeLane(_vclToLeft,10,FALSE)
              end
          else
              begin
                aFrontCar.ChangeLane(_vclToRight,10,FALSE)
              end

          end
        // Front cars return to its former lane.
        else
          begin
          if((Result.Distance < ChangingLaneDistance)and (Result.Distance>0)) then
              if number1 = 2 then
              begin
                aFrontCar.ChangeLane(_vclToLeft,200,FALSE)
              end
              else
                aFrontCar.ChangeLane(_vclToRight,200,FALSE)
          end
        end

        // No cars front.
    else
        begin
        Result.Exists := FALSE;
        Result.Distance := 0.0;
        end;
    end;
end.
